
// p4fliconv: high resolution interlaced FLI converter utility
// Copyright (C) 2007-2008 Istvan Varga <istvanv@users.sourceforge.net>
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//
// The Plus/4 program files generated by this utility are not covered by the
// GNU General Public License, and can be used, modified, and distributed
// without any restrictions.

#ifndef P4FLICONV_DITHER_HPP
#define P4FLICONV_DITHER_HPP

#include "p4fliconv.hpp"

namespace Plus4FLIConv {

  extern const int  ditherTable_FloydSteinberg[13];
  extern const int  ditherTable_Jarvis[13];
  extern const int  ditherTable_Stucki[13];
  extern const int  ditherTable_Sierra2[13];

  extern const int  ditherTable_Bayer[4096];
  extern const int  ditherTable[4096];

  static inline bool ditherPixelValue_Bayer(long xc, long yc, double n)
  {
    return (n >= ((double(ditherTable_Bayer[((yc & 63L) << 6)
                                            | (xc & 63L)]) + 0.5) / 4096.0));
  }

  static inline bool ditherPixelValue(long xc, long yc, double n)
  {
    return (n >= ((double(ditherTable[((yc & 63L) << 6)
                                      | (xc & 63L)]) + 0.5) / 4096.0));
  }

  static inline bool ditherPixelValue(long xc, long yc, double n,
                                      const int *ditherTable_)
  {
    return (n >= ((double(ditherTable_[((yc & 63L) << 6)
                                       | (xc & 63L)]) + 0.5) / 4096.0));
  }

}       // namespace Plus4FLIConv

#endif  // P4FLICONV_DITHER_HPP

